#pragma once

#include <BWAPI.h>
#include "CBase.h"
#include "CUnit.h"
#include "Macro.h"
#include "Util.h"

using namespace BWAPI;

class CheckItem
{
public:
	static void checkBases();
	static void checkEvoChamber();
	static void checkGas();
	static void checkHive();
	static void checkHydraDen();
	static void checkHydralisk();
	static void checkMacroHatch();
	static void checkMutalisk();
	static void checkNewBase();
	static void checkQueensNest();
	static void checkSpawningPool();
	static void checkSpire();
	static void checkSpore();
	static void checkSunken();
	static void checkSupply();
	static void checkWorkers();
	static void checkUltraCavern();
	static void checkUltralisk();
	static void checkZergling();

	static void checkHydraRange();
	static void checkHydraSpeed();
	static void checkLair();
	static void checkLingSpeed();
	static void checkOverlordSpeed();

	static void checkGroundArmor();
	static void checkMissile();
};